// libanthyconv.so Anthy Ѥplugin
//
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <malloc.h>
#include "anthy.xpm"
#include "heke.h"
#include "anthyconv.h"

static RKMap *hira_map,*kata_map;
static atom_t A_extend_segment,A_shorten_segment;

static void init_key_atom()
{
    A_extend_segment = get_atom_by_name("anthy_extend_segment");
    A_shorten_segment = get_atom_by_name("anthy_shorten_segment");

    bind_str_to_atom("C-o",A_extend_segment);
    bind_str_to_atom("S-Right",A_extend_segment);
    bind_str_to_atom("S-Left",A_shorten_segment);
    bind_str_to_atom("C-i",A_shorten_segment);
}


bool init_conv()
{
    // ⤷ǼԤ顢δؿfalse֤
    AnthyConv *anthy = new AnthyConv();
    init_key_atom();
    if ( anthy_init()){
	return false;
    }
    register_kkconv(anthy);
  
    hira_map = new RKMap(hiragana);
    kata_map = new RKMap(katakana);
    return true;
}

void initAnthy()
{
    init_conv();
}

AnthyConv::~AnthyConv()
{
    // λν
}

KKContext *AnthyConv::createContext(XimIC *ic)
{
    // ϥƥȤ롣
    AnthyContext *anthy;
    anthy = new AnthyContext(this,ic);
    return anthy;
}

char **AnthyConv::getIcon()
{
    // toolbar˽ФƤ륢֤
    return anthy_xpm;
}

void AnthyConv::onPushIcon()
{
    // toolbarβ줿˸ƤФ롣
}

/////////////////////////////////////
//
AnthyContext::AnthyContext(AnthyConv *,XimIC *ic) : KKContext(ic)
{
    ac = anthy_create_context();
    mRkConv = new RKConv();
    mRkConv->select_map(hira_map);
    erase_jstring(&mText);
    mMode = INPUT_MODE_RAW;
    mCurSeg = 0;
}

AnthyContext::~AnthyContext()
{
    // ȤʤʤäƥȤ
    anthy_release_context(ac);
}

void AnthyContext::OnUpdatePe(pe_stat *pe)
{
    // preedit˸ƤФ롣
    pe->clear();
    jstring_t::iterator it;
    if ( mMode != INPUT_MODE_HIRA ){
	return ;
    }
    if ( mConvStat == ANTHY_INPUT ){
	pe->new_segment(PE_UNDERLINE);
	for ( it = mText.begin() ; it != mText.end() ; it++){
	    pe->push_cchar(*it);
	}
	char buf[10];
	if (mRkConv->get_pending_char(buf,10)){
	    int i;
	    for ( i = 0 ; i < (int)strlen(buf); i++){
		pe->push_cchar(buf[i]);
	    }
	}
    }else{/*ANTHY_COMPOSE*/
	std::vector<candidates>::iterator jt;
	int i;
	for ( jt = mSeg.begin() , i = 0; jt != mSeg.end() ; jt++,i++){
	    candidates *cands;
	    cands = &(*jt);
	    if ( i == mCurSeg ){
		pe->new_segment(PE_UNDERLINE|PE_REVERSE);
		if ( cands->opCount >3 ){
		    pe->cands = cands;
		}
	    }else{
		pe->new_segment(PE_UNDERLINE);
	    }
	    for ( it = cands->cands[cands->nth].begin() ;
		  it != cands->cands[cands->nth].end() ; it++){
		pe->push_cchar(*it);
	    }
	}
    }
}

int AnthyContext::push_key(keyState *e)
{
    if ( mMode == INPUT_MODE_RAW || mMode == INPUT_MODE_WIDE){
	return proc_raw_mode(e);
    }
    if ( mMode == INPUT_MODE_HIRA ){
	int r;
	if ( mConvStat == ANTHY_INPUT ){
	    r = proc_input_mode(e);
	}else{
	    r = proc_compose_mode(e);
	}
	update_preedit();
	return r;
    }
    return COMMIT_RAW;
}

int AnthyContext::proc_compose_mode(keyState *e)
{
    if ( e->is_bind_to(A_go_right) && mCurSeg < (int)mSeg.size()-1 ){
        set_current_segment(mCurSeg +1);
	return DO_NOTHING;
    }
    if ( e->is_bind_to(A_go_left) && mCurSeg >0 ){
        set_current_segment(mCurSeg -1);
	return DO_NOTHING;
    }
    if ( e->is_bind_to(A_go_line_end)){
	set_current_segment(mSeg.size()-1);
	return DO_NOTHING;
    }
    if ( e->is_bind_to(A_go_line_head)){
	set_current_segment(0);
	return DO_NOTHING;
    }
    if ( e->is_bind_to(A_cancel)){
        mConvStat = ANTHY_INPUT;
        return DO_NOTHING;
    }
    if ( e->is_bind_to(A_extend_segment)){
        anthy_resize_segment(ac,mCurSeg,1);
        get_anthy_stat(mCurSeg);
	return DO_NOTHING;
    }
    if ( e->is_bind_to(A_shorten_segment)){
        anthy_resize_segment(ac,mCurSeg,-1);
        get_anthy_stat(mCurSeg);
	return DO_NOTHING;
    }
    if ( e->is_bind_to(A_commit) || e->is_bind_to(A_return)){
        do_commit();
        mConvStat = ANTHY_INPUT;
        flush();
	return DO_NOTHING;
    }
    candidates *cands = &(mSeg[mCurSeg]);
    int r = cands->proc_key(e,0);
    if ( r & CAND_EXHAUST ){
	cands->nth = 0;
	cands->opCount = 0;
    }
    if ( r & CAND_COMMIT ){
	do_commit();
	mConvStat = ANTHY_INPUT;
	flush();
	return PUSHKEY_AGAIN;
    }
    return DO_NOTHING;
}

void AnthyContext::do_commit()
{
    jstring_t *s=new jstring_t();
    int i;
    for ( i = 0 ; i < (int)mSeg.size() ; i++){
        append_jstring(s,&mSeg[i].cands[mSeg[i].nth]);
	anthy_commit_segment(ac,i,mSeg[i].nth);
    }
    commit_jstring(s);
}

int AnthyContext::proc_input_mode(keyState *e)
{
    /*⡼ڤؤܥ*/
    if ( e->is_bind_to(A_latin_mode) || e->is_bind_to(A_hira_mode)){
	jstring_t *s=new jstring_t();
	append_jstring(s,&mText);
	commit_jstring(s);
	flush();
        if ( e->is_bind_to(A_latin_mode)){
            mMode = INPUT_MODE_RAW;
        }
	return UPDATE_MODE;
    }
    /*ʸǤʤ̤*/
    if ( !mText.size() && !mRkConv->is_pending()){
	int cc;
	cc = e->char_code();
        if ( cc== 0 || cc == 0x20){
            return COMMIT_RAW;
        }
    }
    /*Ѵܥ Ѵ*/
    if ( e->is_bind_to(A_do_conv) && mText.size()){
	char *p;
	p = jstring_to_str(&mText);
	anthy_set_string(ac,p);
	free(p);
	mConvStat = ANTHY_COMPOSE;
	get_anthy_stat(0);
	return DO_NOTHING;
    }
    /*󥻥ܥ*/
    if ( e->is_bind_to(A_cancel)){
	flush();
	return DO_NOTHING;
    }
    /*Хåڡ*/
    if ( e->is_bind_to(A_delete_back)){
	if (! mRkConv->back_space()){
	    if ( mText.size()){
		mText.pop_back();
		return DO_NOTHING;
	    }else{
		return COMMIT_RAW;
	    }
	}
	return DO_NOTHING;
    }
    /*ꤹ*/
    if ( e->is_bind_to(A_commit) || e->char_code() == 0){
	jstring_t *s;
	s = new jstring_t();
	append_jstring(s,&mText);
	commit_jstring(s);
	flush();
	return DO_NOTHING;
    }
    if (!mRkConv->push_key(e->char_code())){
	mText.push_back(e->char_code());
    }
    cchar ch;
    ch = mRkConv->get_cchar();
    while(ch){
	mText.push_back(ch);
	ch = mRkConv->get_cchar();
    }
    return DO_NOTHING;
}

void AnthyContext::flush()
{
    mRkConv->flush();
    erase_jstring(&mText);
}

int AnthyContext::proc_raw_mode(keyState *e)
{
    if ( e->is_bind_to(A_hira_mode)){
	mMode = INPUT_MODE_HIRA;
	mConvStat = ANTHY_INPUT;
	return UPDATE_MODE;
    }
    if ( mMode == INPUT_MODE_WIDE ){
	return COMMIT_WIDE;
    }
    return COMMIT_RAW;
}

int AnthyContext::get_anthy_stat(int n)
{
    char buf[1000];
    anthy_conv_stat cs;
    anthy_get_stat(ac,&cs);
    int i,j,p;
    p = mSeg.size() - n;
    for ( i = 0 ; i < p ; i++){
        mSeg.pop_back();
    }
    for ( i = n ; i < cs.nr_segment ; i++){
        candidates cands;
        anthy_segment_stat ss;
        anthy_get_segment_stat(ac,i,&ss);
        for ( j = 0 ; j < ss.nr_candidate ; j++){
            jstring_t s;
            anthy_get_segment(ac,i,j,buf,1000);
            str_to_jstring(&s,buf);
            cands.cands.push_back(s);
        }
        cands.nth = 0;
        mSeg.push_back(cands);
    }
    set_current_segment(n);
    return 0;
}

int AnthyContext::get_mode()
{
    return mMode;
}

void AnthyContext::set_mode(int m)
{
    if ( mMode == m){
	return ;
    }
    mMode = m;

    flush();
    if ( m == INPUT_MODE_HIRA ){
	mConvStat = ANTHY_INPUT;
	mRkConv->select_map(hira_map);
    }
    if ( m == INPUT_MODE_KATA ){
	mRkConv->select_map(kata_map);
    }
    update_preedit();
}

void AnthyContext::set_current_segment(int s)
{
    mCurSeg = s;
    mSeg[s].opCount = 0;
}

jstring_t *AnthyContext::clear()
{
    // ƥȤ
    return 0;
}
/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 */
